/*
** Copyright (C) 1995, 1996 Open Software Foundation, Inc.
** All Rights Reserved (DCE-Web). Please see COPYRIGHT.DCEWEB
** from the source distribution.
*/

#include "wand.h"

extern pthread_mutex_t getpwnam_m;

/* Return allocated copy of a URI, made canonical. */
/* If h != NULL use halloc routines otherwise NEW */
char *
canon(h_t *h, char *p)
{
    int i;
    char *save;
    char *r, c;
    char *user = NULL;

    /* Move past protocol, as in "http:" */
    if ((p[0] == 'h' || p[0] == 'H') &&
        (p[1] == 't' || p[1] == 'T') &&
        (p[2] == 't' || p[2] == 'T') &&
        (p[3] == 'p' || p[3] == 'P') &&
        (p[4] == ':'))
	p += 5;
    else
	/* XXX same for DCE-http? See OT 111. */
	;

    /* Move past "//host:port". */
    if (p[0] == '/' && p[1] == '/') {
	for (p += 2; *p && *p != '/'; p++)
	    continue;
	if (*p == '\0')
	    return (h) ? hstrdup(h, "/") : dce_strdup("/");
    }

    /* Move past "~foo" or "/~foo". */
    if (p[0] == '~' || (p[0] == '/' && p[1] == '~')) {
        save = p;
	for (p += 1; *p && *p != '/'; p++)
	    continue;
        c = *p;
	*p = '\0';
	i = strlen(save);
	if (*save == '/')
	    r = user = (h) ? halloc(h, i+1) : NEW(char, i+1);
	else {
	    r = user = (h) ? halloc(h, 1+i+1) : NEW(char, 1+i+1);
	    *r++ = '/';
        }
	strcpy(r, save);
	if (c == '\0')
	    return user;
	/* Move past trailing '/' */
	*p++ = '/';
    }

    i = strlen(p);
    if (*p == '/')
        r = save = (h) ? halloc(h, i+1) : NEW(char, i+1);
    else {
        r = save = (h) ? halloc(h, 1+i+1) : NEW(char, 1+i+1);
	*r++ = '/';
    }

    while (*p) {
	if (*p == '/') {
	    for (*r++ = '/'; *p == '/'; p++)
		continue;
	    if (*p == '\0')
		break;
	}

	while (*p && *p != '/')
	    *r++ = *p++;

	/* Did we just copy over "/.." with next char about to be a "/" ? */
	if (r[-1] == '.' && r[-2] == '.' && r[-3] == '/') {
	    if (r == &save[3])
		/* Only have "/.." so just reset to beginning. */
		r = &save[1];
	    else
		/* Have "/x...x/.." so move back over the slash. */
		for (r -= 3; r > save && r[-1] != '/'; r--)
		    continue;

	    /* Now pointing at a slash so don't copy one. */
	    while (*p == '/')
	        p++;
	}
	/* Did we just copy over "/." with next char about to be a "/" ? */
	else if (r[-1] == '.' && r[-2] == '/') {
	    if (r == &save[2])
		/* Only have "/." so just reset to beginning. */
		r = &save[1];
	    else
	        /* back up to '.' */
	        r -= 1;	

	    /* Now pointing at a slash so don't copy one. */
	    while (*p == '/')
	        p++;
	}
    }
    *r = '\0';
    if (user) {
      i = strlen(user) + strlen(save);
      r = (h)? halloc(h, i+1) : NEW(char, i+1);
      sprintf(r, "%s%s", user, save);
      if (!h) DISPOSE(user);
      if (!h) DISPOSE(save);
      save = r;
    }
    return save;
}

/*
** check for bad symbolic links according to policy, returns:
**  0 on a good link
** -1 on a bad link
** -2 on an error
*/
static int
check_links(char *p1, struct stat Sb1, link_checktype_t check, char *p2)
{
    struct stat Sb, Sb2;
    char *q, *end;
    
    switch (check) {
      default:
        break;
      case CL_NOCHECK:
        /*
         * The fastest easiest thing,
         * don't check for sym link correctness
         */
        return 0;
      case CL_SUBTREE:
        /*
         * Check if the owner of path is the same
         * as the owner of p2, works well in userdirs
         * so that users in their homedirs can only link
         * to something else they own.
         */
        if (stat(p2, &Sb2) == -1) {
            dce_svc_printf(WAND_S_UNIX_FUNCTION_WARNING_MSG, "stat", errno);
            return -2;
        }
        return (Sb1.st_uid == Sb2.st_uid) ? 0 : -1;
      case CL_LINKOBJ:
        /*
         * Loop throught the path checking if each rdn is a symlink,
         * if so make sure the owner of the link is the same as the
         * owner of the target of the link
         */
        end=p1+strlen(p1);
        for (q=strchr(p1+1,'/'); q && q<=end; *q='/', q=strchr(q+1, '/')) {
            *q='\0';
            if (lstat(p1, &Sb) == -1) {
                dce_svc_printf(WAND_S_UNIX_FUNCTION_WARNING_MSG,
                               "lstat", errno);
                return -2;
            }
            if (S_ISLNK(Sb.st_mode)) {
                if (stat(p1, &Sb2) == -1) {
                    dce_svc_printf(WAND_S_UNIX_FUNCTION_WARNING_MSG,
                                   "stat", errno);
                    return -2;
                }
                if (Sb.st_uid != Sb2.st_uid)
                    return -1;
            }
        }
        if (lstat(p1, &Sb) == -1) {
            dce_svc_printf(WAND_S_UNIX_FUNCTION_WARNING_MSG, "lstat", errno);
            return -2;
        }
        if (S_ISLNK(Sb.st_mode)) {
            if (stat(p1, &Sb2) == -1) {
                dce_svc_printf(WAND_S_UNIX_FUNCTION_WARNING_MSG,
                               "stat", errno);
                return -2;
            }
            if (Sb.st_uid != Sb2.st_uid)
                return -1;
        }
        return 0;
    }

    return -2;                                    /* bad value */
}

/*
 * uri is something like "melman/Images/foo.gif"
 * userdir is probably "public_html"
 * return allocated string of "/u1/melman/public_html/Images/foo.gif"
 */
static char *
expand_user_dir(h_t *h, char *uri, char *userdir, char *shortcut, char **eod)
{
    char          *user;
    char          *home;
    char          *p, *q, *r;
    char          tmp;
    struct passwd *pw;
    int           i;
    
    /* get user from uri and move uri past the username */
    for (user=uri; *uri!='/' && *uri!='\0'; uri++)
        ;
    /* if more to uri after user, temporarily change until after getpwnam */
    if ((tmp=*uri))
        *uri++='\0';
        
    /* Note: home alloced in shortcut case, not in getpwnam case */
    if (shortcut) {
        /* just prepent shortcut to username, perhaps add / in between */
        /* shortcut is meant to be something like /homes/ */
        i=strlen(user)+strlen(shortcut)+1;
        p = home = (h) ? halloc(h, i) : NEW(char, i);
        for (q=shortcut; *q; *p++=*q++)           /* copy in shortcut */
            ;
        if (p[-1] != '/')                         /* add / only if needed */
            *p++='/';
        for (q=user; *q; *p++=*q++)               /* copy in user */
            ;
        *p='\0';                                  /* trailing \0 */
    }
    else {
        /* get home directory of user, getpwnam not threadsafe */
        thread_mutex_lock(&getpwnam_m);
        pw=getpwnam(user);
        if (!pw || !pw->pw_dir) {
            thread_mutex_unlock(&getpwnam_m);
            return NULL;
        }
        home=pw->pw_dir;                          /* note still not alloc'ed */
    }

    /* restore / if needed */
    if (tmp)
        uri[-1]=tmp;

    /* now build <home>/<userdir>/<uri> */
    i=strlen(home)+1+strlen(userdir)+1+strlen(uri)+1;
    p = r = (h) ? halloc(h, i) : NEW(char, i);
    for (q=home; *q; *p++=*q++)                   /* copy in home */
        ;
    thread_mutex_unlock(&getpwnam_m);             /* release mutex asap */
    if (shortcut && !h)                           /* we alloced in this case */
        DISPOSE(home);
    if (p[-1] != '/')                             /* add / only if needed */
        *p++='/';
    if (userdir) {
        for (q=userdir; *q; *p++=*q++)            /* copy in userdir */
            ;
        if (p[-1] != '/')                         /* add / only if needed */
            *p++='/';
    }
    if (eod)                                      /* record end of home dir */
        *eod=p;
    for (q=uri; *q; *p++=*q++)                    /* copy in remaining uri */
        ;
    *p='\0';                                      /* trailing \0 */

    return r;
}

static void
find_uri_path_int(h_t *h,
                  char *uri,
                  char **canon_uri,
                  char **query,
                  char **path,
                  char **inuserdir,
                  error_status_t *st)
{
    char *p;
    int i;

    *st=dceweb_st_ok;

    /* get the canonical uri */
    if (*canon_uri == NULL)
        *canon_uri = canon(h, uri);

    p=*canon_uri;

    /* sanity check and move past initial / */ 
    if (*p++ != '/') {
        *st=dceweb_st_internalerror;
        return;
    }
    
    /* everything after ? is a query string */
    *query=strchr(p, '?');
    
    /* decode the URI, tracking query if present */
    if ((i=uri_decode(p, query)) < 0) {
        *st=dceweb_st_badrequest;
        return;
    }

    /* remove query for a bit */
    if (*query)
        **query = '\0';
    
    /*
     * Three cases here:
     *   1. If there is a ~ it's a user directory
     *   2. URI was just "/" so must copy ACL_ROOT to make things work below
     *   3. Any other uri just copy it
     * After all three *path must end with a "/"
     */
    if (*p == '~')
        /* userdir and userhome are global from config file */
        *path = expand_user_dir(h, ++p, userdir, userhome, inuserdir);
    else if (*p == '\0')                          /* we removed / above */
        *path = (h) ? hstrdup(h, ACL_ROOT) : dce_strdup(ACL_ROOT);
    else                           
        *path = (h) ? hstrdup(h, p) : dce_strdup(p);
    
    /* if there was a query, put it back in canon_uri for CGIs */
    if (*query) {
        **query = '?';
        (*query)++;                                 /* move past the ? */
    }

    if (!*path)
        *st=dceweb_st_notfound;

    return;
}

static void
find_uri_path_h(h_t *h, char **path, char **inuserdir, error_status_t *st)
{
    find_uri_path_int(h,h->uri,&h->canon_uri,&h->query,path,inuserdir,st);
}

void
find_uri_path(char *uri,
              char **canon_uri,
              char **query,
              char **path,
              char **inuserdir,
              error_status_t *st)
{
    find_uri_path_int(NULL, uri, canon_uri, query, path, inuserdir,st);
}


/* Look in the uri field of the handle.
 * Relative uri's begin with a / 
 * Since we don't support proxy, ALL uri's should be relative
 *
 * Only set *st in error case!
 */
void
find_uri_h(h_t *h, char **location, error_status_t *st)
{
    char *u, *p, *path, *end;
    int  i, r;
    error_status_t ost;
    char *inuserdir=NULL;
    char *home=NULL;
    int link_ok;
    
    *st=dceweb_st_ok;

    find_uri_path_h(h, &path, &inuserdir, &ost);
    if (ost != dceweb_st_ok) {
        *st=ost;
        return;
    }
    
    /* Find a good path.
     * We have the name of a file or
     * the name of a directory or
     * the name of a script followed by other rdns or
     * a bad name
     */     
    if ((r=stat(path, &h->stat)) == -1) {
        if (errno == EACCES) {
            *st=dceweb_st_forbidden;
            return;
        }
        if (errno != ENOENT && errno != ENOTDIR) {
            *st=dceweb_st_internalerror;
            return;
        }
    }
    
    /* if request was in a userdir we'll need home either now or later */
    if (inuserdir) {
        int l=inuserdir-path;

        home=halloc(h, l+1);
        strncpy(home, path, l);
        home[l]='\0';
    }

    if (!r) {
        /* Check that symlinks are valid */
        if (inuserdir && cl_userdir)
            link_ok=check_links(path, h->stat, cl_userdir, home);
        else if (cl_docroot)
            link_ok=check_links(path, h->stat, cl_docroot, docroot);
        else
            link_ok=0;

        if (link_ok) {
            *st=(link_ok==-1)? dceweb_st_badsymlink : dceweb_st_internalerror;
            return;
        }
    }

    /* if a dir look for index.html */
    if (!r && (S_ISDIR(h->stat.st_mode))) {
        /*
         * URL's to directories must end with a / or relative URLs 
         * referenced on same page will be wrong.  Return re-direct
         * to the same uri as they used with / appended
         */
        i=strlen(h->uri);
        if (h->uri[i-1]!='/') {
            static char L[]="Location: ";
            char *p=h->query;
            char *q;
            int j;

            if (p)
                *--p='\0';                        /* remove the query */
            j=3*strlen(h->canon_uri)+1;
            q=halloc(h, j);
            j=uri_encode(h->canon_uri, q);

            j+=STRLEN(L)+strlen(h->uribase)+3+1;
            *location=halloc(h, j);
            sprintf(*location, "%s%s%s/\r\n", L, h->uribase, q);
            if (p)
                *p='?';                           /* put the query back */
            *st=dceweb_st_movedtemp;
            return;
        }
        else {
            char *path2;
            struct stat Sb2;

            /* append index.html? */
	    if (EQ(path, ACL_ROOT))
                path2=hstrdup(h, dirindexstr);
	    else
                path2=hglue_names(h, path, 0, dirindexstr);
            if (!stat(path2, &Sb2)) {             /* have index.html */
                path=path2;
                stat(path, &h->stat);           /* can't assign struct stat */
            }
        }
    }
    
    /* find the end */
    for (p=path; *p; p++)
        ;
    end=p;

    /* If this path is a dir or a file we're done, otherwise
     * parse the name backward to find the script.
     */
    if (r || !(S_ISDIR(h->stat.st_mode) || S_ISREG(h->stat.st_mode))) {
        boolean32 changed;
        char      *old;
        
        changed=0;
        do {
            old=p;
            p=strrchr(path, '/');

            if (changed)
                *--old='/';
            
            if ((!p && p!=end) || p==path) {
                *st=dceweb_st_notfound;
                return;
            }
            
            else {
                *p++='\0';
                changed=1;
            }
            if ((i=stat(path, &h->stat)) == -1) {
                if (errno == EACCES) {
                    *st=dceweb_st_forbidden;
                    return;
                }
                else if (errno != ENOENT && errno != ENOTDIR) {
                    *st=dceweb_st_internalerror;
                    return;
                }
            }
        } while ((i==-1) || !S_ISREG(h->stat.st_mode));

        /* Check that symlinks are valid */
        if (inuserdir && cl_userdir)
            link_ok=check_links(path, h->stat, cl_userdir, home);
        else if (cl_docroot)
            link_ok=check_links(path, h->stat, cl_docroot, docroot);
        else
            link_ok = 0;

        if (link_ok) {
            *st=(link_ok==-1)? dceweb_st_badsymlink : dceweb_st_internalerror;
            return;
        }
    }

    h->file_path=path;
    h->path_info=p;
    h->path_trans=NULL;
    return;
}
